@@ -3,13 +3,44 @@
 ** Copyright (C) 2000-2001 Nullsoft, Inc.
 ** Author: Justin Frankel
 ** File: httpget.cpp - JNL HTTP GET implementation
-** License: see jnetlib.h
+** License: see License.txt
+**
+** Unicode support by Jim Park -- 08/24/2007
 */
 
 #include "netinc.h"
 #include "util.h"
 #include "httpget.h"
 
+static const char g_errstr_oom[] = "error allocating memory";
+
+template<class T> static int classify_scheme_char(T c)
+{
+  if (('A' <= c && c <= 'Z') || ('a' <= c && c <= 'z')) return 0;
+  return c == '+' || c == '-' || c == '.' ? 1 : -1;
+}
+
+template<class T> static unsigned int get_scheme_length(T*s)
+{
+  if (classify_scheme_char(s[0]) != 0) return 0; // The first character MUST be a-z||A-Z. (RFC 3986 3.1)
+  for (unsigned int i=0;;)
+  {
+    if (s[++i] == ':') return ++i; // Length including ':'
+    if (classify_scheme_char(s[i]) < 0) return 0;
+  }
+}
+
+enum { GLT_ABS=0, GLT_NETREL, GLT_ROOTREL, GLT_PATHREL };
+template<class T> static unsigned int get_location_type(T*s)
+{
+  unsigned int len=get_scheme_length(s);
+  if (len) return GLT_ABS; // "xyz:..."
+  s+=len;
+  // Note: The authority section is not skipped here because it is assumed to not exist without a scheme!
+  if (s[0] != '/') return GLT_PATHREL; // "xyz..." or "./xyz..." (or "" but don't rely on this)
+  if (s[1] == '/') return GLT_NETREL; // "//..." (relative "network-path reference")
+  return GLT_ROOTREL; // "/" or "/xyz..."
+}
 
 JNL_HTTPGet::JNL_HTTPGet(JNL_AsyncDNS *dns, int recvbufsize, char *proxy)
 {
@@ -71,7 +102,7 @@ JNL_HTTPGet::~JNL_HTTPGet()
 }
 
 
-void JNL_HTTPGet::addheader(char *header)
+void JNL_HTTPGet::addheader(const char *header)
 {
   //if (strstr(header,"\r") || strstr(header,"\n")) return;
   if (!m_sendheaders)
@@ -133,7 +164,7 @@ void JNL_HTTPGet::do_encode_mimestr(char *in, char *out)
 }
 
 
-void JNL_HTTPGet::connect(char *url)
+void JNL_HTTPGet::connect(const char *url)
 {
   deinit();
   m_http_url=(char*)malloc(strlen(url)+1);
@@ -173,20 +204,20 @@ void JNL_HTTPGet::connect(char *url)
   char *str=(char*)malloc(sendbufferlen+1024);
   if (!str)
   {
-    seterrstr("error allocating memory");
+    seterrstr(g_errstr_oom);
     m_http_state=-1;    
   }
 
   if (!m_http_proxyhost || !m_http_proxyhost[0])
   {
-    wsprintf(str,"GET %s HTTP/1.0\r\n",m_http_request);
+    wsprintfA(str,"GET %s HTTP/1.0\r\n",m_http_request);
   }
   else
   {
-    wsprintf(str,"GET %s HTTP/1.0\r\n",m_http_url);
+    wsprintfA(str,"GET %s HTTP/1.0\r\n",m_http_url);
   }
 
-  wsprintf(str+strlen(str),"Host: %s\r\n",m_http_host);
+  wsprintfA(str+strlen(str),"Host: %s\r\n",m_http_host);
 
   if (m_http_lpinfo&&m_http_lpinfo[0])
   {
@@ -228,7 +259,7 @@ void JNL_HTTPGet::connect(char *url)
 
 }
 
-static int my_strnicmp(char *b1, char *b2, int l)
+static int my_strnicmp(char *b1, const char *b2, int l)
 {
   while (l-- && *b1 && *b2)
   {
@@ -241,20 +272,20 @@ static int my_strnicmp(char *b1, char *b2, int l)
   return 0;
 }
 
-char *_strstr(char *i, char *s)
+char *_strstr(const char *i, const char *s)
 {
   if (strlen(i)>=strlen(s)) while (i[strlen(s)-1])
   {
     int l=strlen(s)+1;
-    char *ii=i;
-    char *is=s;
+    const char *ii=i;
+    const char *is=s;
     while (--l>0)
     {
       if (*ii != *is) break;
       ii++;
       is++;
     }
-    if (l==0) return i;
+    if (l==0) return const_cast<char*>(i);
     i++;
   }
   return NULL;
@@ -310,20 +341,20 @@ void JNL_HTTPGet::do_parse_url(char *url, char **host, int *port, char **req, ch
 }
 
 
-char *JNL_HTTPGet::getallheaders()
+const char *JNL_HTTPGet::getallheaders()
 { // double null terminated, null delimited list
   if (m_recvheaders) return m_recvheaders;
   else return "\0\0";
 }
 
-char *JNL_HTTPGet::getheader(char *headername)
+const char *JNL_HTTPGet::getheader(const char *headername)
 {
   char *ret=NULL;
   if (strlen(headername)<1||!m_recvheaders) return NULL;
   char *p=m_recvheaders;
   while (*p)
   {
-    if (!my_strnicmp(headername,p,strlen(headername)))
+    if (!my_strnicmp(p,headername,strlen(headername)))
     {
       ret=p+strlen(headername);
       while (*ret == ' ') ret++;
@@ -359,6 +390,11 @@ run_again:
       m_con->recv_line(buf,4095);
       buf[4095]=0;
       m_reply=(char*)malloc(strlen(buf)+1);
+      if (!m_reply)
+      {
+        seterrstr(g_errstr_oom);
+        return m_http_state=-1;
+      }
       strcpy(m_reply,buf);
 
       if (strstr(buf,"200")) m_http_state=2; // proceed to read headers normally
@@ -391,7 +427,58 @@ run_again:
         char *p=buf+9; while (*p== ' ') p++;
         if (*p)
         {
-          connect(p);
+          const char *orgloc=m_http_url, *e;
+          char *newloc=p, *newbuf=0;
+          if (my_strnicmp(newloc, "http://",7))
+          {
+            unsigned int lenorg=0, sep=0, lennew=0;
+            unsigned int newloctype=get_location_type(newloc);
+
+            for (e=orgloc, lenorg=7; e[lenorg]; ++lenorg) // This assumes the original URL starts with "http://"
+            {
+              if (e[lenorg] == '/' || e[lenorg] == '?' || e[lenorg] == '#') // Is it the end of the authority section?
+                break;
+            }
+
+            if (newloctype == GLT_ABS)
+            {
+              // A scheme we don't understand, let connect deal with it.
+            }
+            else
+            {
+              lennew=strlen(newloc);
+              if (newloctype == GLT_NETREL)
+              {
+                lenorg=sizeof("http:")-1; // The new location is "//...", we must prepend the original scheme.
+              }
+              else if (newloctype != GLT_ROOTREL)
+              {
+                ++sep; // We need a separator between the original path and the new path.
+                unsigned int lastsep=lenorg;
+                for (e=orgloc; e[lastsep]; ++lastsep)
+                {
+                  if (e[lastsep] == '/')
+                    lenorg = lastsep;
+                  if (e[lastsep] == '?' || e[lastsep] == '#') // Is it the end of the path section?
+                    break;
+                }
+              }
+
+              p=malloc(lenorg+sep+lennew+1);
+              if (!p)
+              {
+                seterrstr(g_errstr_oom);
+                return m_http_state=-1;
+              }
+              strncpy(p,orgloc,++lenorg), p[lenorg]='\0';
+              if (sep) strcat(p, "/");
+              strcat(p,newloc);
+              newbuf=newloc=p;
+            }
+          }
+
+          connect(newloc);
+          free(newbuf);
           return 0;
         }
       }
@@ -484,7 +571,7 @@ int JNL_HTTPGet::peek_bytes(char *buf, int len)
 
 __int64 JNL_HTTPGet::content_length()
 {
-  char *p=getheader("content-length:");
+  const char *p=getheader("content-length:");
   if (!p) return 0;
   __int64 cl = myatoi64(p);
   if (cl > 0) return cl;
