@@ -478,7 +478,10 @@ void CDialogTemplate::ConvertToRTL() {
       }
     }
     else if (!IS_INTRESOURCE(m_vItems[i]->szClass) && !WinWStrICmpASCII(m_vItems[i]->szClass, "SysTreeView32")) {
-      m_vItems[i]->dwStyle |= TVS_RTLREADING;
+      // we can't have both TVS_RTLREADING and WS_EX_LAYOUTRTL because the text becomes non-LTR
+      // we chose just WS_EX_LAYOUTRTL because just TVS_RTLREADING changes the text only and not the tree itself
+      // https://github.com/qbittorrent/qBittorrent/pull/19929
+      //m_vItems[i]->dwStyle |= TVS_RTLREADING;
       m_vItems[i]->dwExtStyle |= WS_EX_LAYOUTRTL;
       addExStyle = true;
       addExLeftScrollbar = false;
