@@ -3,7 +3,7 @@
  * 
  * This file is a part of NSIS.
  * 
- * Copyright (C) 1999-2019 Nullsoft and Contributors
+ * Copyright (C) 1999-2023 Nullsoft and Contributors
  * 
  * Licensed under the zlib/libpng license (the "License");
  * you may not use this file except in compliance with the License.
@@ -22,7 +22,6 @@
 #include "state.h"
 #include "ui.h"
 #include "lang.h"
-#include "state.h"
 #include "exec.h"
 #include "plugin.h"
 
@@ -96,18 +95,50 @@ EXTERN_C void NSISWinMainNOCRT()
   TCHAR *realcmds;
   TCHAR seekchar=_T(' ');
   TCHAR *cmdline;
+  OSVERSIONINFOEX ovi;
 
   SetErrorMode(SEM_NOOPENFILEERRORBOX | SEM_FAILCRITICALERRORS);
-  g_WinVer = GetVersion() & ~(NSIS_WINVER_WOW64FLAG); // We store a private flag in the build number bits
+
+  // Get the version as reported by Windows
+  if (sizeof(void*) < 8)
+  {
+    *((UINT32*)&ovi.szCSDVersion[0]) = 0; // Zero out SP
+    *((UINT64*)&ovi.wServicePackMajor) = 0; // wServicePackMajor, wSuiteMask and wProductType
+  }
+  ovi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
+  if (!GetVersionEx((OSVERSIONINFO*) &ovi) && sizeof(void*) < 8)
+  {
+    ovi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
+    GetVersionEx((OSVERSIONINFO*) &ovi);
+    if (sizeof(TCHAR) == 2 || ovi.dwPlatformId == VER_PLATFORM_WIN32_NT)
+    {
+      ovi.wProductType = 4; // TODO: For < NT4SP6, look it up in the registry. 4 means not W9x and not VER_NT_*
+      ovi.wServicePackMajor = ovi.szCSDVersion[0] == 'S' ? ovi.szCSDVersion[13] - '0' : 0;
+    }
+  }
+  if (sizeof(TCHAR) == 1 && ovi.dwPlatformId < VER_PLATFORM_WIN32_NT)
+  {
+    ovi.wProductType = 0;
+    ovi.wServicePackMajor = ovi.szCSDVersion[1] >= 'A' ? ovi.szCSDVersion[1] - ('A'-1) : 0; // A, B or C
+  }
+  if (sizeof(void*) < 8 && ovi.dwMajorVersion < 10) // Ideally (sizeof(TCHAR) == 1 && ovi.dwMajorVersion < 5) but the compatibility tab emulates this bug
+  {
+    ovi.dwBuildNumber &= 0xffff; // Remove W9x garbage
+  }
+  // Save the packed version information
+  {
+    UINT32 *p = &g_osinfo.WVBuild;
+    p[0] = ovi.dwBuildNumber;
+    p[1] = MAKELONG(MAKEWORD(ovi.wProductType, ovi.wServicePackMajor), MAKEWORD(ovi.dwMinorVersion, ovi.dwMajorVersion));
+  }
 
   {
     // bug #1125: Don't load modules from the application nor current directory.
     // SetDefaultDllDirectories() allows us to restrict implicitly loaded and 
     // dynamically loaded modules to just %windir%\System32 and directories 
     // added with AddDllDirectory(). This prevents DLL search order attacks (CAPEC-471).
-    DWORD winver = g_WinVer;
     // CoCreateInstance(CLSID_ShellLink, ...) fails on Vista if SetDefaultDllDirectories is called
-    BOOL avoidwinbug = LOWORD(winver) == MAKEWORD(6, 0);
+    BOOL avoidwinbug = IsWinVista();
     if (!avoidwinbug)
     {
       FARPROC fp = myGetProcAddress(MGA_SetDefaultDllDirectories);
@@ -160,7 +191,7 @@ EXTERN_C void NSISWinMainNOCRT()
     // accessing a unsupported view and RegKey* takes care of that by looking at the WOW64 flag.
     FARPROC fp = myGetProcAddress(MGA_IsOS);
     enum { os_wow6432 = 30 };
-    if (fp && ((BOOL(WINAPI*)(UINT))fp)(os_wow6432)) g_WinVer |= NSIS_WINVER_WOW64FLAG;
+    if (fp && ((BOOL(WINAPI*)(UINT))fp)(os_wow6432)) g_osinfo.WVProd |= NSIS_OSINFO_PROD_WOW64FLAG;
   }
 #endif
 
@@ -317,60 +348,69 @@ EXTERN_C void NSISWinMainNOCRT()
     }
     else
     {
-      int x, admin = UserIsAdminGrpMember();
-
-      mystrcat(state_temp_dir,_T("~nsu"));
-      if (admin) mystrcat(state_temp_dir,_T("A")); // Don't lock down the directory used by non-admins
-      mystrcat(state_temp_dir,_T(".tmp"));
-
-      // check if already running from uninstaller temp dir
-      // this prevents recursive uninstaller calls
-      if (!lstrcmpi(state_temp_dir,state_exe_directory))
-        goto end;
-
-      admin ? CreateRestrictedDirectory(state_temp_dir) : CreateNormalDirectory(state_temp_dir);
-      SetCurrentDirectory(state_temp_dir);
-
-      if (!(*state_install_directory))
-        mystrcpy(state_install_directory,state_exe_directory);
+      int admin = UserIsAdminGrpMember();
+      WORD tries; // 0xfffe attempts or bust
+      size_t cchtmpslash = mystrlen(state_temp_dir);
+      LPTSTR unexe = g_usrvars[5], unexecmd = g_usrvars[6];
 
       mystrcpy(g_usrvars[0], realcmds);
-      SET2CHAR(g_usrvars[1], _T("A\0"));
-
-      for (x = 0; x < 26; x ++)
+      if (!(*state_install_directory))
       {
-        static TCHAR buf2[NSIS_MAX_STRLEN];
+        mystrcpy(state_install_directory, state_exe_directory);
+      }
 
-        GetNSISString(buf2,g_header->str_uninstchild); // $TEMP\Un_$1.exe
-
-        DeleteFile(buf2); // clean up after all the other ones if they are there
-
-        if (m_Err) // not done yet
+      for (tries = 0; ++tries != 0;)
+      {
+        DWORD retry = 0, ec;
+retry_un_dir:
+        wsprintf(state_temp_dir + cchtmpslash, _T("~nsu%X.tmp"), tries);
+        GetNSISString(unexe, g_header->str_uninstchild); // '$TEMP\Un.exe'
+        if (admin)
         {
-          // copy file
-          if (CopyFile(state_exe_path,buf2,TRUE))
+          ec = CreateRestrictedDirectory(state_temp_dir);
+        }
+        else
+        {
+          ec = CreateNormalDirectory(state_temp_dir);
+        }
+
+        if (ec)
+        {
+          // Delete previous uninstaller (if it is safe to do so) (Bug #1296)
+          if (!(GetFileAttributes(unexe) & FILE_ATTRIBUTE_REPARSE_POINT) && DeleteFile(unexe))
           {
-            HANDLE hProc;
+            myDelete(state_temp_dir, DEL_DIR);
+            if (!retry++) goto retry_un_dir;
+          }
+        }
+        else
+        {
+          HANDLE hProc;
+          SetCurrentDirectory(state_temp_dir);
 #ifdef NSIS_SUPPORT_MOVEONREBOOT
-            MoveFileOnReboot(buf2,NULL);
+          MoveFileOnReboot(state_temp_dir, NULL);
 #endif
-            GetNSISString(buf2,g_header->str_uninstcmd); // '"$TEMP\Un_$1.exe" $0 _?=$INSTDIR\'
-            hProc=myCreateProcess(buf2);
+          if (CopyFile(state_exe_path, unexe, TRUE))
+          {
+#ifdef NSIS_SUPPORT_MOVEONREBOOT
+            MoveFileOnReboot(unexe, NULL);
+#endif
+            GetNSISString(unexecmd, g_header->str_uninstcmd); // '"$TEMP\Un.exe" $0 _?=$INSTDIR\'
+            hProc = myCreateProcess(unexecmd);
             if (hProc)
             {
               CloseHandle(hProc);
-              // success
-              m_Err = 0;
+              m_Err = 0; // Success
+            }
+            else if (!retry++ && !file_exists(unexe))
+            {
+              // Another instance deleted us between CopyFile and CreateProcess
+              goto retry_un_dir;
             }
           }
+          break; // We called CreateProcess; success or failure, we are done.
         }
-        (*(((NSIS_STRING *)g_usrvars)[1]))++;
       }
-
-#ifdef NSIS_SUPPORT_MOVEONREBOOT
-      MoveFileOnReboot(state_temp_dir,NULL);
-#endif
-
       goto end;
     }
   }
