@@ -3,7 +3,7 @@
  * 
  * This file is a part of NSIS.
  * 
- * Copyright (C) 1999-2019 Nullsoft and Contributors
+ * Copyright (C) 1999-2023 Nullsoft and Contributors
  * 
  * Licensed under the zlib/libpng license (the "License");
  * you may not use this file except in compliance with the License.
@@ -62,20 +62,16 @@ const UINT32 g_restrictedacl[] = {
   0x10000000, // ACCESS_ALLOWED_ACE:ACCESS_MASK: GENERIC_ALL
   0x00000201, 0x05000000, 0x00000020, 0x00000220, // ACCESS_ALLOWED_ACE:SID (BUILTIN\Administrators) NOTE: GetAdminGrpSid() relies on this being the first SID in the ACL
   0x00140300, // ACCESS_ALLOWED_ACE:ACE_HEADER (ACCESS_ALLOWED_ACE_TYPE, CONTAINER_INHERIT_ACE|OBJECT_INHERIT_ACE)
-  0x00130041, // ACCESS_ALLOWED_ACE:ACCESS_MASK: DELETE|READ_CONTROL|SYNCHRONIZE|FILE_DELETE_CHILD|FILE_LIST_DIRECTORY
+  0x001200c1, // ACCESS_ALLOWED_ACE:ACCESS_MASK: SYNCHRONIZE|READ_CONTROL|FILE_LIST_DIRECTORY|FILE_DELETE_CHILD|FILE_READ_ATTRIBUTES
   0x00000101, 0x01000000, 0x00000000 // ACCESS_ALLOWED_ACE:SID (WORLD\Everyone)
 };
 
 DWORD NSISCALL CreateRestrictedDirectory(LPCTSTR path)
 {
-  const SECURITY_INFORMATION si = OWNER_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION|DACL_SECURITY_INFORMATION|PROTECTED_DACL_SECURITY_INFORMATION;
   PSID admingrpsid = GetAdminGrpSid();
   SECURITY_DESCRIPTOR sd = { 1, 0, SE_DACL_PRESENT, admingrpsid, admingrpsid, NULL, GetAdminGrpAcl() };
   SECURITY_ATTRIBUTES sa = { sizeof(SECURITY_ATTRIBUTES), &sd, FALSE };
-  DWORD ec = CreateDirectory(path, &sa) ? ERROR_SUCCESS : GetLastError();
-  if (ERROR_ALREADY_EXISTS == ec)
-    ec = SetFileSecurity(path, si, &sd) ? ERROR_SUCCESS : GetLastError();
-  return ec;
+  return CreateDirectory(path, &sa) ? ERROR_SUCCESS : GetLastError();
 }
 DWORD NSISCALL CreateNormalDirectory(LPCTSTR path)
 {
@@ -394,7 +390,7 @@ int NSISCALL is_valid_instpath(TCHAR *s)
     return 0;
 
   // must be called after skip_root or AllowRootDirInstall won't work.
-  // validate_filename removes trailing blackslashes and so converts
+  // validate_filename removes trailing backslashes and so converts
   // "C:\" to "C:" which is not a valid directory. skip_root returns
   // NULL for "C:" so the above test returns 0.
   // validate_filename is called so directories such as "C:\ " will
@@ -771,13 +767,38 @@ TCHAR * NSISCALL mystrcat(TCHAR *out, const TCHAR *concat)
   return lstrcat(out, concat);
 }
 
+int StrWideToACP(LPCWSTR Src, char* Dst, int DstCap)
+{
+  return WideCharToMultiByte(CP_ACP, 0, Src, -1, Dst, DstCap, NULL, NULL);
+}
+
+#ifndef UNICODE
+void strcpyWideToT(TCHAR *out, LPCWSTR in)
+{
+  StrWideToACP(in, out, NSIS_MAX_STRLEN);
+}
+#endif
+
+#if !defined(_WIN64) && !defined(UNICODE)
+HRESULT ComIIDFromString(LPCTSTR str, IID*out)
+{
+  WCHAR buf[130];
+  signed char i;
+  for (i = 0; i >= 0; ++i)
+    if (!(buf[i] = str[i]))
+      return IIDFromString(buf, out);
+  return E_FAIL;
+}
+#endif
+
 TCHAR ps_tmpbuf[NSIS_MAX_STRLEN*2];
 
 const TCHAR SYSREGKEY[]   = _T("Software\\Microsoft\\Windows\\CurrentVersion");
 const TCHAR QUICKLAUNCH[] = _T("\\Microsoft\\Internet Explorer\\Quick Launch");
 
-typedef HRESULT (__stdcall * PFNSHGETFOLDERPATH)(HWND, int, HANDLE, DWORD, LPTSTR);
+typedef HRESULT (WINAPI* PFNSHGETFOLDERPATH)(HWND, int, HANDLE, DWORD, LPTSTR);
 extern void *g_SHGetFolderPath;
+typedef LPITEMIDLIST (WINAPI* PFNSHCLONESPECIALIDLIST)(HWND, int, BOOL);
 
 // Based on Dave Laundon's simplified process_string
 // The string actually has a lot of different data encoded into it.  This
@@ -827,7 +848,7 @@ TCHAR * NSISCALL GetNSISString(TCHAR *outbuf, int strtab)
         LPITEMIDLIST idl;
 
         int x = 2;
-        DWORD ver = sizeof(void*) > 4 ? MAKEWORD(5, 2) : g_WinVer; // We only care about 95/98 vs ME/NT4+
+        BOOL isWin9598 = IsWin9598();
         /*
         SHGetFolderPath as provided by shfolder.dll is used to get special folders
         unless the installer is running on Windows 95/98. For 95/98 shfolder.dll is
@@ -845,8 +866,7 @@ TCHAR * NSISCALL GetNSISString(TCHAR *outbuf, int strtab)
         */
         BOOL use_shfolder =
           // Use shfolder if not on 95/98
-          !((ver & 0x80000000) && (LOWORD(ver) != MAKEWORD(4,90))) ||
-
+          !isWin9598 ||
           // Unless the Application Data or Documents folder is requested
           (
             (fldrs[3] == CSIDL_COMMON_APPDATA) ||
@@ -880,16 +900,22 @@ TCHAR * NSISCALL GetNSISString(TCHAR *outbuf, int strtab)
 
         while (x--)
         {
+          PFNSHCLONESPECIALIDLIST SHCSILFunc;
+          int idandflags = fldrs[x] & ~0x40, create = idandflags & CSIDL_FLAG_CREATE;
+
           if (g_SHGetFolderPath && use_shfolder)
           {
             PFNSHGETFOLDERPATH SHGetFolderPathFunc = (PFNSHGETFOLDERPATH) g_SHGetFolderPath;
-            if (!SHGetFolderPathFunc(g_hwnd, fldrs[x], NULL, SHGFP_TYPE_CURRENT, out))
+            if (!SHGetFolderPathFunc(g_hwnd, idandflags, NULL, SHGFP_TYPE_CURRENT, out))
             {
               break;
             }
           }
-            
-          if (!SHGetSpecialFolderLocation(g_hwnd, fldrs[x], &idl))
+
+          // This function exists in 95 & NT4 but was undocumented back then so we have to import it by ordinal.
+          SHCSILFunc = (PFNSHCLONESPECIALIDLIST) myGetProcAddress(MGA_SHCloneSpecialIDList);
+          idl = SHCSILFunc(g_hwnd, LOBYTE(idandflags), create);
+          if (idl)
           {
             BOOL res = SHGetPathFromIDList(idl, out);
             CoTaskMemFree(idl);
@@ -1142,10 +1168,14 @@ struct MGA_FUNC MGA_FUNCS[] = {
 #ifndef _WIN64
   {"KERNEL32", "GetDiskFreeSpaceExW"},
   {"KERNEL32", "GetUserDefaultUILanguage"},
+#endif
+#if !defined(_WIN64) || defined(_M_IA64)
   {"ADVAPI32", "RegDeleteKeyExW"},
 #endif
   {"ADVAPI32", "InitiateShutdownW"},
+  {"SHELL32", "SHGetKnownFolderPath"},
   {"SHELL32", (CHAR*) 680}, // IsUserAnAdmin
+  {"SHELL32", (CHAR*) 89}, // SHCloneSpecialIDList
 #ifndef _WIN64
   {"SHLWAPI", (CHAR*) 437}, // IsOS
 #endif
@@ -1163,7 +1193,9 @@ struct MGA_FUNC MGA_FUNCS[] = {
   {"KERNEL32", "GetUserDefaultUILanguage"},
   {"ADVAPI32", "RegDeleteKeyExA"},
   {"ADVAPI32", "InitiateShutdownA"},
+  {"SHELL32", "SHGetKnownFolderPath"},
   {"SHELL32", (CHAR*) 680}, // IsUserAnAdmin
+  {"SHELL32", (CHAR*) 89}, // SHCloneSpecialIDList
 #ifndef _WIN64
   {"SHLWAPI", (CHAR*) 437}, // IsOS
 #endif
@@ -1234,7 +1266,7 @@ void * NSISCALL NSISGetProcAddress(HANDLE dllHandle, TCHAR* funcName)
 {
 #ifdef _UNICODE
   char ansiName[256];
-  if (WideCharToMultiByte(CP_ACP, 0, funcName, -1, ansiName, 256, NULL, NULL) != 0)
+  if (StrWideToACP(funcName, ansiName, 256) != 0)
     return GetProcAddress(dllHandle, ansiName);
   return NULL;
 #else
